#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#define MIN2(y,z)        ((y)<(z))?(y):(z)
#define MIN3(x,y,z)      MIN2((x),MIN2((y),(z)))
#define MIN4(w,x,y,z)    MIN2((w),MIN3((x),(y),(z)))


// Newick: (((One:0.2,Two:0.3):0.3,(Three:0.5,Four:0.3):0.2):0.3,Five:0.7):0.0;

// Takes a tree in newick format, builds an internal "tree" structure
// generates calls to other programs with correct weights



typedef struct tree_node {
  struct tree_node* lc;
  struct tree_node* rc;
  struct tree_node* par;
  char* seqname; 
  char* aligned;
  int leafs;
  int mynumout;
  int alignlen;
  float distup;
} node;


int Ksize;
node** leaflist = 0;
int numleafs = 0;
int missing_dist = 0;
node** outgroups;
float* outdists;  
float* table;
int numouts;


char* dna_alpha = "ACGT";
char* valid_alpha = "ACGTN-";
char* DNA_PRINT;
char* DNA_LET;
char* NUM_ONES;

void init_consts() {
  int i;
  outgroups = (node**) malloc (sizeof(node*) * numleafs);
  outdists = (float*) malloc (sizeof(float) * numleafs);  
  DNA_LET = (char*) malloc (sizeof(char) * 0x10);
  DNA_PRINT = (char*) malloc (sizeof(char) * 0x10);
  NUM_ONES = (char*) malloc (sizeof(char) * 0x10);

  for (i=0; i < 0x10; i++) {
    NUM_ONES[i] = DNA_LET[i] = DNA_PRINT[i] = -1;
  }

  DNA_LET[0] = 4;
  DNA_LET[1] = 0;
  DNA_LET[2] = 1;
  DNA_LET[4] = 2;
  DNA_LET[8] = 3;
  DNA_PRINT[0] = 'N';
  DNA_PRINT[1] = 'A';
  DNA_PRINT[2] = 'C';
  DNA_PRINT[4] = 'G';
  DNA_PRINT[8] = 'T';
  DNA_PRINT[1|2] = 'M';
  DNA_PRINT[1|4] = 'R';
  DNA_PRINT[1|8] = 'W';
  DNA_PRINT[2|4] = 'S';
  DNA_PRINT[2|8] = 'Y';
  DNA_PRINT[4|8] = 'K';
  DNA_PRINT[1|2|4] = 'V';
  DNA_PRINT[1|2|8] = 'H';
  DNA_PRINT[1|4|8] = 'D';
  DNA_PRINT[2|4|8] = 'B';
  DNA_PRINT[1|2|4|8] = 'X';
  NUM_ONES[0] = 1;
  NUM_ONES[1] = 1;
  NUM_ONES[2] = 1;
  NUM_ONES[4] = 1;
  NUM_ONES[8] = 1;
  NUM_ONES[1|2] = 2;
  NUM_ONES[1|4] = 2;
  NUM_ONES[1|8] = 2;
  NUM_ONES[2|4] = 2;
  NUM_ONES[2|8] = 2;
  NUM_ONES[4|8] = 2;
  NUM_ONES[1|2|4] = 3;
  NUM_ONES[1|2|8] = 3;
  NUM_ONES[1|4|8] = 3;
  NUM_ONES[2|4|8] = 3;
  NUM_ONES[1|2|4|8] = 4;
}


node* mk_node() {
  node* res = (node*)malloc(sizeof(node));
  res->leafs = 0;
  res->lc = res->rc = res->par = 0;
  res->seqname = 0;
  res->aligned = 0;
  res->mynumout = -1;
  res->distup = -1.0;
  return res;
}

void clean_node(node* res) {
  //  res->leafs = 0;
  //  res->lc = res->rc = res->par = 0;
  //  free (res->seqname);
  //  res->seqname = 0;
  free(res->aligned);
  res->aligned = 0;
  //  res->mynumout = -1;
  //  res->distup = -1.0;
}

void cleanupTree(node* root) {
  if (!root)
    return;
  cleanupTree(root->lc);
  cleanupTree(root->rc);
  clean_node(root);
}


void parse_error(char* str, int m, char* symb) {
  fprintf(stderr, "Error in input %s\n", str+m);
  fprintf(stderr, "Was expecting %s\n", symb);
  exit(1);
}

int mystrchr(char* str, int ch) {
  int k = (int)strchr(str,ch);
  return (k)? k : 9999999;
}

int read_name (char* str, node* myroot) {
  int n = (int) MIN3(mystrchr(str, ','), mystrchr(str,':'), 
		     mystrchr(str,')'));

  n = n - (int)str;
  //  fprintf(stderr, "n=%d\n", n);
  if (n < 0) {
    parse_error(str, 0, "%s{,:)}");
  }

  numleafs++;
  leaflist = realloc (leaflist, sizeof(node*) * numleafs);
  leaflist[numleafs-1] = myroot;
  myroot->seqname = (char*) malloc(n+1);
  strncpy(myroot->seqname, str, n);
  myroot->seqname[n] = 0;
  return n;
}

int read_dist (char* str, node* myroot) {
  float d;
  int n;
  if (!sscanf(str, "%f%n", &d, &n))
    parse_error(str, 0, "%f");
  myroot->distup = d;
  return n;
}

int parse_tree(char* tree_str, node* myroot) {
  int moved = 0;
  //  fprintf(stderr, "Proc %s\n", tree_str);
  if (tree_str[0] != '(') {
    myroot->leafs = 1;
    moved += read_name(tree_str, myroot);
    if (tree_str[moved] == ':') {
      //      fprintf(stderr, "Proc2 %s\n", tree_str+moved);
      moved += 1;
      moved += read_dist(tree_str+moved, myroot);
      //      fprintf(stderr, "Proc3 %s\n", tree_str+moved);
    }
    else {
      missing_dist = 1;
    }
  }
  else {
    moved = 1;
    myroot->lc = mk_node();
    myroot->lc->par = myroot;
    moved += parse_tree(tree_str+1, myroot->lc);
    if (tree_str[moved] != ',') {
      parse_error(tree_str, moved, ",");
    }
    myroot->rc = mk_node();
    myroot->rc->par = myroot;
    moved +=1;
    moved += parse_tree(tree_str+moved, myroot->rc);
    if (tree_str[moved] != ')') {
      parse_error(tree_str, moved, ")");
    }
    moved += 1;
    if (tree_str[moved] == ':') {
      moved += 1;
      moved += read_dist(tree_str+moved, myroot);
    }
    else {
      missing_dist = 1;
    }
    myroot->leafs = myroot->lc->leafs + myroot->rc->leafs;
  }
  //  fprintf(stderr, " done: %s", tree_str+moved);
  return moved;
}

float get_down_groups (node* mynode, float dsofar) {
  dsofar += (mynode->distup*mynode->leafs);
  if (mynode->lc || mynode->rc) {
    assert (mynode->lc && mynode->rc);
    return get_down_groups(mynode->lc, dsofar) + 
      get_down_groups(mynode->rc, dsofar);
  }
  outgroups[numouts] = mynode;
  mynode->mynumout = numouts;
  outdists[numouts] = 1/dsofar;
  numouts++;
  return 1/dsofar;
}

//get me my outgroups
float get_outgroups (node* mynode, float dsofar) {
  node* oc = 0;
  float tsize = 0;
  dsofar += (mynode->distup*(numleafs - mynode->leafs));
  if (mynode->par) {
    if (mynode->par->lc == mynode)
      oc = mynode->par->rc;
    else if (mynode->par->rc == mynode)
      oc = mynode->par->lc;
    else {
      fprintf(stderr, "error in tree structure\n");
      exit(1);
    }
    tsize += get_outgroups(mynode->par, dsofar);
    tsize += get_down_groups(oc, dsofar);
  }
  return tsize;
}

void genParsimony (node* myroot, node* lr, node* rr) {
  int i;

  if (!lr || !rr) {
    lr = (lr)? lr : rr;
    myroot->aligned = malloc (sizeof(char) * lr->alignlen);
    myroot->alignlen = lr->alignlen;

    for (i = 0; i < lr->alignlen; i++) {
      myroot->aligned[i] = lr->aligned[i];
    }
    return;
  }
  myroot->aligned = malloc (sizeof(char) * lr->alignlen);
  myroot->alignlen = lr->alignlen;
  for (i=0; i < lr->alignlen; i++) {
    assert (lr->alignlen == rr->alignlen);
    if (lr->aligned[i] & rr->aligned[i])
      myroot->aligned[i] = lr->aligned[i] & rr->aligned[i];
    else
      myroot->aligned[i] = lr->aligned[i] | rr->aligned[i];
  }
}

node* find_align_node (node* myroot) {
  int nlen, i;
  float ttree, test = 0;
  node *fc, *sc, *lr, *rr;

  //  fprintf(stderr, "into \n");
  if (myroot ->leafs == 1) {// leaf
    //    fprintf(stderr,"if %s\n", myroot->seqname);
    if (myroot->aligned)
      return myroot;
    else
      return 0;
  }
  else {
    //    fprintf(stderr, "else \n");
    lr = find_align_node(myroot->lc);
    rr = find_align_node(myroot->rc);

    if (!myroot->seqname) {
      nlen = strlen (myroot->lc->seqname) + strlen(myroot->rc->seqname) + 2;
      myroot->seqname = malloc(nlen);
      //    set name    
      if (strcmp(myroot->lc->seqname, myroot->rc->seqname) < 0) {
	fc = myroot->lc;
	sc = myroot->rc;
      }
      else {
	fc = myroot->rc;
	sc = myroot->lc;      
      }
      strcpy(myroot->seqname, fc->seqname);
      strcat(myroot->seqname, "_");
      strcat(myroot->seqname, sc->seqname);
    }
  }
  
  if (lr || rr)
    genParsimony(myroot, lr, rr);

  if (lr && rr) {
    return myroot;
  }
  else if (lr) {
    //    fprintf(stderr,"warning 1\n");
    return lr;
  }
  else if (rr) {
    //    fprintf(stderr,"warning 2\n");
    return rr;
  }
  return 0;
}



int read_align(FILE* input, node **target) {
  char* res = (char*) malloc(sizeof(char)*1);
  int i, ressize = 1, numread=0, namelen; 
  char temp[1024];
  char currchar, bitchar, *tt;

  if (feof(input)) {
    return 0;
  }


  fgets(temp, 255, input);
  if (temp[0] != '>') {
    fprintf(stderr, "File is not in FASTA format!!\n");
    exit(1);
  }
  *(strchr(temp, '\n')) = 0;

  currchar = fgetc(input);

  while ((currchar != '>') && (currchar != EOF)) {
    if (!isspace(currchar)) {
      currchar = toupper(currchar);
      bitchar = 0;
      if (!(tt = strchr(valid_alpha, currchar)) || (tt - valid_alpha > 3)) {
	//        fprintf(stderr, "Warning: %d:%c skipped'\n", numread,currchar);
        currchar = 'N';
	res[numread] = 0;
      }
      else {
	res[numread] = (1 << (int)(tt-valid_alpha));
      }
      numread++;
      if (numread >= ressize) {
        res=(char*)realloc(res, sizeof(char)*(ressize*=2));
      }
    }
    currchar = fgetc(input);
  }
  //find node
  for (i = 0; i < numleafs; i++) {
    //    fprintf(stderr,"check %s  %s\n", temp+1, leaflist[i]->seqname);
    if (!strncmp(temp+1, leaflist[i]->seqname, strlen(leaflist[i]->seqname))) {
      if (!leaflist[i]->aligned) {
	//	fprintf(stderr,"SET %s\n", leaflist[i]->seqname);
	if (target)
	  *target = leaflist[i];
	leaflist[i]->aligned = res;
	leaflist[i]->alignlen = numread;
      }
      else {
	//	fprintf(stderr,"%s was already set\n", leaflist[i]->seqname);
      }
    }
  }

  if (currchar == '>') {
    ungetc(currchar, input);
    return 1;
  }
  return 0;


}

int read_align_file (FILE* input, int k) {
  int i, j=0;
  for (i=0; i < k; i++) {
    if (!read_align(input,0)) {
      j = 1;
      break;
    }
  }
  if (j) {
    assert (i==0 || i==(k-1));
  }
  return i;
}

int maxima(node* a, float** distrib, int pos) {
  int i, max = -1;
  for (i = 0; i < 4; i++) {
    if (a->aligned[pos] & (1 << i)) {
      if (max < 0)
	max = i;
      else if (distrib[i][pos] > distrib[max][pos])
	max = i;
    }
  }
  return (1 << max);
}

int numvariants = 1;

void doWord(node*a, float ** distrib, int i) {
  char stack[31];
  int sp = 0;
  int j;
  int word;
  float val;

  numvariants *= NUM_ONES[a->aligned[i+Ksize-1]];

  stack [0] = -1;

  while (sp >= 0) {
    //    fprintf(stderr, "sp = %d Ksize = %d\n", sp, Ksize);
    while(stack[sp] < 4) {
      //find next match
      stack[sp]++;
      if (a->aligned[i+sp] & (1 << stack[sp])) {
	stack[++sp] = -1;
	break;
      }
    }
    if (sp == Ksize) {
      word = 0;
      for (j=0; j < Ksize; j++) {//add, update distrib
	word |= (stack[j] << (j*2));
      }
      val = table[word];
      for (j=0; j < Ksize; j++) {//add, update distrib
	distrib[stack[j]][i+j] += (val/numvariants);
	//	fprintf(stderr,"update: %d:%d %f\n", i+j, stack[j], val/numvariants);
      }      
    }
    if ((sp == Ksize) || (stack[sp] >= 4)) {
      sp--;
    }
  }

  numvariants /= NUM_ONES[a->aligned[i]];
}

void genConsensus (node* a) {
  char* newres = malloc (sizeof(char) * a->alignlen), *tt;
  float* distrib[4];
  int i,j;

  for (i=0; i < 4; i++) {
    distrib[i] = (float*) malloc (sizeof(float) * a->alignlen);
    for (j=0; j < a->alignlen; j++) 
      distrib[i][j] = 0;
  }


  //setup numvariants
  numvariants = 1;
  for (i=0; i < Ksize-1; i++) {
    numvariants *= NUM_ONES[a->aligned[i]];
  }
  for (i=0; i < a->alignlen-Ksize+1; i++) {
    //take word
    if (!(i%1000)) {
      fprintf(stderr,"working %d/%d\n",i/1000,a->alignlen/1000);
    }
    doWord(a, distrib, i);
    //sum over all variants

  }
  for (i=0; i < a->alignlen; i++) {
    if (DNA_LET[a->aligned[i]] > 0)
      newres[i] = a->aligned[i];
    else {
      newres[i] = maxima(a, distrib, i);
    }
  }
  for (i=0; i< 4; i++) {
    free(distrib[i]);
  }

  tt= a->aligned;
  a->aligned = newres;
  free (tt);
}

void printConsensus (node* a) {
  int i;
  fprintf(stdout, ">%s:%d ", a->seqname, a->alignlen);
  for (i=0; i < a->alignlen; i++) {
    if (!(i%60))
      fprintf(stdout, "\n");
    //    fprintf(stdout, "%d:[%x]%c", i+1,a->aligned[i],DNA_PRINT[a->aligned[i]]);
    fprintf(stdout, "%c", DNA_PRINT[a->aligned[i]]);

  }
  fprintf(stdout, "\n");
  for (i=0; i < (1<< (Ksize*2)); i++) {
    //    fprintf(stdout, "%f\n", table[i]);
  }
}

void read_table(char* tabfile) {
  FILE* f = fopen (tabfile, "r");
  if (!f) {
    fprintf(stderr, "Error opening table\n");
    exit(2);
  }
  fread(&Ksize, sizeof(int), 1, f);
  table = (float*) malloc (sizeof(float) * (1 << (2*Ksize)));
  fread(table, sizeof(float), (1 << 2 * Ksize), f);
}

void lagan2newick (char* target, char* source, int size) {
  int i, pos=0, w=0;
  char temp[16537];
  int len = strlen(source); 
  for (i=0; i < len; i++) { //compress whitespace
    if (w && isspace(source[i])) {
      fprintf(stderr, "here\n");
      continue;
    }
    else {
      if (w == 1 && (source[i] == '(' || source[i] == ')'))
	temp[pos-1] = source[i];
      else
	temp[pos++] = source[i];
    }

    if (isspace(source[i]))
      w = 1;
    else if (source[i] == ')')
      w = 2;
    else
      w = 0;
  }
  temp[pos] = 0;
  fprintf(stderr, "temp = %s\n", temp);


  len = pos;
  pos = 1;
  target[0] = temp[0];
    for (i=1; i < len; i++) {
    if ((temp[i] == '(' && temp[i-1] == ')') ||
	(temp[i] == '(' && temp[i-1] != '(') ||
	(temp[i-1] == ')' && temp[i] != ')')) 
      {
	target[pos++] = ',';
      }
    if (isspace(temp[i]))
      target[pos++] = ',';
    else
      target[pos++] = temp[i];
    if (len + pos >= size) {
      fprintf(stderr, "Wow, that's a biggie... readjust the constants (string tree len)\n");
      exit(1);
    }
  }
  target[pos] = ';';
  target[pos+1] = 0;
  return;
}

int main(int argc, char** argv) {
  char string_tree[16537]; //noone will ever need more :)))
  int moved, i, k;
  float ttree, test;
  FILE* alfile;
  node *root = mk_node();
  node *align_node;
  
  if (argc != 5) {
    fprintf(stderr, "Usage: consensus alignment.mfa table \"lagan tree\" #seqsperfile\n");
    exit(2);
  }

  //  fprintf(stderr, "Parsed tree\n");
  if (strlen(argv[3]) > 16536) {
    fprintf(stderr, "Wow, that's a biggie... readjust the constants (string tree len)\n");
    return 1;
  }

  lagan2newick(string_tree, argv[3], 16536); 
  fprintf(stderr,"Cnversion, was:\n%s\nnow: %s\n", argv[3], string_tree); 

  moved = parse_tree(string_tree, root);
  if (string_tree[moved] != ';') {
    parse_error(string_tree, moved, ";");
  }



  init_consts();
  read_table(argv[2]);
  fprintf(stderr, "read k = %d\n", Ksize);

  k = atoi(argv[4]);

  // Read alignment
  alfile = fopen(argv[1], "r");
  if (!alfile) {
    fprintf(stderr, "COULDN'T OPEN ALIGNMENT\n");
    exit (2);
  }

  while (read_align_file(alfile, k)) {

    align_node = find_align_node(root);
    genConsensus(align_node);
    printConsensus(align_node);
    cleanupTree(root);

  }
  return 0;
}
