#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#define MIN2(y,z)        ((y)<(z))?(y):(z)
#define MIN3(x,y,z)      MIN2((x),MIN2((y),(z)))
#define MIN4(w,x,y,z)    MIN2((w),MIN3((x),(y),(z)))


// Newick: (((One:0.2,Two:0.3):0.3,(Three:0.5,Four:0.3):0.2):0.3,Five:0.7):0.0;

// Takes a tree in newick format, builds an internal "tree" structure
// generates calls to other programs with correct weights



typedef struct sequence {
  char* seqname; 
  char* aligned;
  int alignlen;
  int mynum;
} seq;


int Ksize; 

seq** outgroups;
float* outdists;  
float* table;
int numouts;


char* dna_alpha = "ACGT";
char* valid_alpha = "ACGTN-";
char* DNA_PRINT;
char* DNA_LET;
char* NUM_ONES;

void init_consts() {
  int i;
  table = (float*) malloc (sizeof(float) * (1 << (2*Ksize)));
  outgroups = (seq**) malloc (sizeof(seq*) * numouts);
  outdists = (float*) malloc (sizeof(float) * numouts);  
  DNA_LET = (char*) malloc (sizeof(char) * 0x10);
  DNA_PRINT = (char*) malloc (sizeof(char) * 0x10);
  NUM_ONES = (char*) malloc (sizeof(char) * 0x10);

  for (i=0; i < 0x10; i++) {
    NUM_ONES[i] = DNA_LET[i] = DNA_PRINT[i] = -1;
  }

  DNA_LET[1] = 0;
  DNA_LET[2] = 1;
  DNA_LET[4] = 2;
  DNA_LET[8] = 3;
  DNA_PRINT[0] = 'N';
  DNA_PRINT[1] = 'A';
  DNA_PRINT[2] = 'C';
  DNA_PRINT[4] = 'G';
  DNA_PRINT[8] = 'T';
  DNA_PRINT[1|2] = 'M';
  DNA_PRINT[1|4] = 'R';
  DNA_PRINT[1|8] = 'W';
  DNA_PRINT[2|4] = 'S';
  DNA_PRINT[2|8] = 'Y';
  DNA_PRINT[4|8] = 'K';
  DNA_PRINT[1|2|4] = 'V';
  DNA_PRINT[1|2|8] = 'H';
  DNA_PRINT[1|4|8] = 'D';
  DNA_PRINT[2|4|8] = 'B';
  DNA_PRINT[1|2|4|8] = 'X';
  NUM_ONES[0] = 0;
  NUM_ONES[1] = 1;
  NUM_ONES[2] = 1;
  NUM_ONES[4] = 1;
  NUM_ONES[8] = 1;
  NUM_ONES[1|2] = 2;
  NUM_ONES[1|4] = 2;
  NUM_ONES[1|8] = 2;
  NUM_ONES[2|4] = 2;
  NUM_ONES[2|8] = 2;
  NUM_ONES[4|8] = 2;
  NUM_ONES[1|2|4] = 3;
  NUM_ONES[1|2|8] = 3;
  NUM_ONES[1|4|8] = 3;
  NUM_ONES[2|4|8] = 3;
  NUM_ONES[1|2|4|8] = 4;
}


seq* mk_seq() {
  seq* res = (seq*)malloc(sizeof(seq));
  res->seqname = 0;
  res->aligned = 0;
  res->mynum = -1;
  return res;
}

int read_align(FILE* input, seq *target) {
  char* res = (char*) malloc(sizeof(char)*1);
  int i, ressize = 1, numread=0; //N at 1st letter
  char temp[1024];
  char currchar, bitchar, *tt;

  if (feof(input)) {
    fprintf(stderr, "2COULDN'T READ ALIGNMENT\n");
    exit (2);
  }


  fgets(temp, 255, input);
  fprintf(stderr, temp);
  if (temp[0] != '>') {
    fprintf(stderr, "File is not in FASTA format!!\n");
    exit(1);
  }
  *(strchr(temp, '\n')) = 0;

  currchar = fgetc(input);

  while ((currchar != '>') && (currchar != EOF)) {
    if (!isspace(currchar)) {
      currchar = toupper(currchar);
      bitchar = 0;
      if (!(tt = strchr(valid_alpha, currchar)) || (tt - valid_alpha > 3)) {
	//        fprintf(stderr, "Warning: %d:%c skipped'\n", numread,currchar);
        currchar = 'N';
	res[numread] = 0;
      }
      else {
	res[numread] = (1 << (int)(tt-valid_alpha));
      }
      numread++;
      if (numread >= ressize) {
        res=(char*)realloc(res, sizeof(char)*(ressize*=2));
      }
    }
    currchar = fgetc(input);
  }

  target->aligned = res;
  target->alignlen = numread;

  if (currchar == '>') {
    ungetc(currchar, input);
    return 1;
  }
  return 0;


}

void cleanseq (seq* t) {
  free(t->aligned);
  t->alignlen = 0;
}

void read_sequences(int argc, char**argv) {
  char* filename;
  FILE* input;
  seq* myn;
  int i, j, kmer, breaker;
  int zz;

  for (i=2; i < argc; i+=2) {
    filename = argv[i];
    myn = 0;
    if (!(input = fopen (filename, "r"))) {
      fprintf(stderr, "COULDN'T OPEN SEQ %d %s\n",i,argv[i]);
      exit (2);
    }
    fprintf(stderr, "fn:%s\n", filename);
    do {
      myn= outgroups[i/2-1];
      myn->mynum = i/2-1;
      zz = read_align(input,myn);
      fprintf(stderr, "zz %d\n",zz);
      if (myn && myn->alignlen >= Ksize) {
	assert (myn->mynum >=0);
	assert (outgroups[myn->mynum] == myn);
	//process
	//init k-mer
	
	kmer = 0; 
	breaker = 0;
	for (j = 0; j < Ksize-1; j++) {
	  if (DNA_LET[myn->aligned[j]] >= 0)
	    kmer |= DNA_LET[myn->aligned[j]] << (2*j);
	  else 
	    breaker = j+1;
	}
	while (j < myn->alignlen) {
	  if (DNA_LET[myn->aligned[j]] >= 0)
	    kmer |= DNA_LET[myn->aligned[j]] << (2*(Ksize-1));
	  else
	    breaker = Ksize;
	  //update table
	  if (!breaker) {
	    //	    fprintf(stderr, "add %d %d %f\n", kmer, myn->mynum, outdists[myn->mynum]);
	    table[kmer] += outdists[myn->mynum];
	  }
	  else {
	    //	    fprintf(stderr, "skip %d\n",kmer);
	    breaker--;
	  }
	  //	  fprintf(stderr,"%c: t[%x] = %f\n", DNA_PRINT[myn->aligned[j]], kmer, table[kmer]);
	  //update k-mer
	  kmer = kmer >> 2;
	  j++;
	}
      }
      free(myn->aligned);
      myn = 0;
    } while (zz)
	;
    fclose(input);
  }
}

void dump_table() {
  fprintf(stderr, "DUMPING\n");
  fwrite(&Ksize, sizeof(int), 1, stdout);
  fwrite(table, sizeof(float), (1 << 2 * Ksize), stdout);
}

int main(int argc, char** argv) {
  char string_tree[16537]; //noone will ever need more :)))
  int moved, i;
  float ttree, test;
  
  //  fprintf(stderr, "Parsed tree\n");
  if (argc < 4) {
    fprintf(stderr, "Usage: mktabl Ksize seq1 seq1w [seq2 seq2w].... > table_file\n");
    exit(2);
  }
  Ksize = atoi(argv[1]);
  numouts = (argc-1)/2;
  init_consts();


  for (i=0; i<numouts; i++) {
    outgroups[i] = mk_seq();
  }

  //  ttree = get_outgroups(align_node, 0);
  //  fprintf(stdout, "ALIGN %s %s RES %s OUTS", align_node->lc->seqname, 
  // 	  align_node->rc->seqname, align_node->seqname);
  for (i=0; i< numouts; i++) {
    outdists[i] = atof(argv[2*(i+1)+1]);
  }
  //  for (i=0; i< numouts; i++) {
  //    fprintf(stdout, " %s %f", outgroups[i]->seqname, outdists[i]);
  //    test += outdists[i];
  //  }
  //  fprintf(stdout, "\n");
  
  read_sequences(argc, argv);
  dump_table();
  return 0;
}
